﻿using Eingabeverwaltung.Metamodell;
using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.IO;
using System.Linq;

namespace Ausgabeverwaltung
{
    /// <summary>
    /// Outputs in File "Automatische_Bewertung.txt" when OutPutMode is "file", which is standard.
    /// Outputs in the "OutPut" string when OutPutMode is "string".
    /// </summary>
    public class Ausgabeverwalter
    {
        private string OutPutMode;
        private string OutPut;
        public Ausgabeverwalter()
        {
            OutPutMode = "file";
            OutPut = "";
        }

        public Ausgabeverwalter(string OutPutMode)
        {
            if (OutPutMode == "file" || OutPutMode == "string")
                this.OutPutMode = OutPutMode;
            else throw new Exception("Unknown OutPutMode set for Ausgabeverwalter: " + OutPutMode);
            OutPut = "";
        }

        /// <summary>
        /// Returns OutPut-string. Warning: OutPut-string is empty if OutPutMode is not "string".
        /// </summary>
        /// <returns></returns>
        public string GetOutPut() => OutPut;

        public void Ausgabe(Container con, ImmutableList<Bewertung.IAssessmentResult> bewertung, int BewertungRound)
        {
            if (bewertung[0] is Bewertung.AssessmentResult)
                Ausgabe(con, bewertung.Cast<Bewertung.AssessmentResult>().ToList().ToImmutableList(), BewertungRound);
            else throw new Exception("Can only output AssessmentResults but got "+bewertung[0].GetType());
        }

        public void Ausgabe(Container con, ImmutableList<Bewertung.AssessmentResult> Bewertung, int BewertungRound)
        {
            (double,double) gesamtBewertung = (0.0,0.0);
            foreach (var ar in Bewertung)
            {
                gesamtBewertung.Item1 += ar.assessment;
                gesamtBewertung.Item2 += ar.MaxPt();
            }
            gesamtBewertung.Item1 = Math.Round(gesamtBewertung.Item1, BewertungRound);
            string result = "Gesamtbewertung: " + gesamtBewertung.Item1 + " von " + gesamtBewertung.Item2 + "\r\n\r\n";
            result += "Bewertung nach Teilaufgaben:\r\n";
            foreach(var ar in Bewertung)
            {
                if (ar.task.name == "main task") result += "Hauptaufgabe: "; else result += "Lemma " + ar.task.name + ": ";
                result += Math.Round(ar.assessment,BewertungRound) + " von " + ar.MaxPt() + "\r\n";
                foreach (var ipa in ar.confIndAssessment)
                    result += "  "+ ipa.partID+": "+ Math.Round(ipa.Assessment, BewertungRound) + " von " + ipa.maxPt + "\r\n";
            }
            result += "\r\nFehler nach Lemmata:\r\n\r\n"+ printAllErrors(con);
            if (OutPutMode == "file")
            {
                File.WriteAllText("Automatische_Bewertung.txt", result);
                Console.WriteLine("Ausgabe in Automatische_Bewertung.txt.");
            }
            else if (OutPutMode == "string") OutPut = result;
            else throw new Exception("Unknown OutPutMode set in Ausgabeverwalter: " + OutPutMode);
        }

        public string printAllErrors(Container con)
        {
            string str = "";
            str += printAllErrors(con.Proof.proof);
            if (str != "") str = "Hauptbeweis\r\n" + str;
            foreach(Lemma lemma in con.Proof.lemmata)
            {
                string lemmastr = "";
                if (lemma.errors.Count > 0) lemmastr += ErrorMessage(lemma.errors, false, -1, -1);
                lemmastr += printAllErrors(lemma.proof);
                if (lemmastr != "")
                {
                    str += "Lemma " + lemma.name;
                    if (lemma.usefulFor.Count > 0)
                    {
                        str += " geht ein in Teilaufgaben: ";
                        foreach (TaskAssignment ta in lemma.usefulFor)
                            str += ta.ToString() + " ";
                    } else str += " ohne Bewertung ";
                    str += "\r\n";
                    str += lemmastr;
                }
            }
            if (str == "") str = NoErrors();
            return str;
        }

        private string printAllErrors(SingleProof sp)
        {
            if (sp is Transformation tr) return printAllErrors(tr);
            else if (sp is Induction ind) return printAllErrors(ind);
            else throw new Exception("Found a proof, that is neither Transformation nor Induction!");
        }

        private string printAllErrors(Induction ind)
        {
            string str = "";
            if (ind.errors.Count > 0) str += ErrorMessage(ind.errors, true, "", false, -1, -1);
            foreach(InductionHypothesis ih in ind.indHyps)
                if (ih.errors.Count > 0) str += ErrorMessage(ih.errors, true, ih.name, false, -1, -1);
            foreach (InductionCase ic in ind.indCases)
            {
                string IndPart = "Fall ";
                if (ic.Constructor != null) IndPart += ic.Constructor.name;
                else IndPart += ic.eq.leftTree.ToString() + " = " + ic.eq.rightTree.ToString();
                string casestr = "";
                if (ic.errors.Count > 0) casestr += ErrorMessage(ic.errors, true, IndPart, false, -1, -1);
                casestr += printAllErrors(ic.proof,true,IndPart);
                if (casestr != "") str += casestr;
            }
            return str;
        }

        private string printAllErrors(Transformation tr) => printAllErrors(tr, false, "");
        private string printAllErrors(Transformation tr, bool ind, string indPart)
        {
            string str = "";
            if (tr.errors.Count > 0) str += ErrorMessage(tr.errors,ind,indPart,true,-1,-1); 
            for(int tsi = 0; tsi < tr.transformationSteps.Count; tsi++)
            {
                TransformationStep ts = tr.transformationSteps[tsi];
                if (ts.errors.Count > 0) str += ErrorMessage(ts.errors, ind, indPart, true, (tsi + 1), -1);
                for(int subi = 0; subi <  ts.substitutions.Count; subi++)
                {
                    Substitution sub = ts.substitutions[subi];
                    if (sub.errors.Count > 0) str += ErrorMessage(sub.errors, ind, indPart, true, (tsi + 1), (subi + 1)); 
                }
            }
            return str;
        }

        public string NoErrors() => "No Errors.";

        public string ErrorMessage(List<Error> errors, int ts, int sub) => ErrorMessage(errors, true, ts, sub);
        public string ErrorMessage(List<Error> errors, bool tr, int ts, int sub) => ErrorMessage(errors, false, "", tr, ts, sub);
        /// <param name="errors"></param>
        /// <param name="ind">Induction or not?</param>
        /// <param name="indPart">if ind: identifying name of part.</param>
        /// <param name="tr"> true -> Transformation; false -> Lemma</param>
        /// <param name="ts">Number of transformationstep</param>
        /// <param name="sub">Number of substitution</param>
        /// <returns></returns>
        public string ErrorMessage(List<Error> errors, bool ind, string indPart, bool tr, int ts, int sub)
        {
            string str = "Errors in ";
            if (ind)
                if (indPart == "") str += "Induction";
                else               str += indPart;
            if (!tr && !ind) str += "Lemma";
            if (tr)
            { 
                if      (ind      ) str += ", ";
                if      (ts  == -1) str += "Transformation";
                else if (sub == -1) str += "TransformationStep " + ts;
                else                str += "TransformationStep " + ts + ", Substitution " + sub;
            }
            str += ":\r\n";
            foreach (Error err in errors)
                str += "    " + err.errorCode + " " + err.Output + "\r\n";
            return str;
        }
    }
}