﻿using Eingabeverwaltung.Metamodell;
using System;
using System.Collections.Generic;
using System.Collections.Immutable;

namespace Bewertung {
    public class Bewerter : AbstractBewerter
    {
        private Container con;
        public Bewerter(Container con) : base(con) { this.con = con; } 

        public Bewerter(Container con, string config) : base(con,config) { this.con = con; }

        public override void Bewerte()
        {
            initializeBewertung();
            bool NoErrors() 
            {
                Ausgabeverwaltung.Ausgabeverwalter aus = new Ausgabeverwaltung.Ausgabeverwalter();
                return aus.printAllErrors(con) == aus.NoErrors();
            }
            if (NoErrors())
            {   // if NoErrors, don't calculate assessment, instead give maxPt.
                // Cause: due to wrong configuration of minSteps calculation of assessment could give less than maxPt for a correct proof.
                foreach(AssessmentObject aO in bewertung)
                {
                    aO.assessment = aO.task.config.MaxPt();
                    if (aO.task.config is InductionConfig indConf)
                        foreach (InductionPartAssessment ipa in aO.confIndAssessment)
                            ipa.Assessment = ipa.maxPt;
                }
                return;
            }
            foreach(AssessmentObject aO in bewertung)
            {
                if (aO.usefulLemmata.Count == 0) continue;
                List<Lemma> usefulLemmata = new List<Lemma>(aO.usefulLemmata);
                Lemma mainLemma = usefulLemmata[0];
                if (aO.task.config is InductionConfig indConf && mainLemma.proof is Induction ind) // Assessment with induction config ( is 0.0 if mainLemma.proof is Transformation)
                {
                    usefulLemmata.RemoveAt(0); // do not regular assessment on main lemma later
                    if (HasIndErrors(ind)) continue; // assessment is 0.0
                    foreach (Lemma lemma in usefulLemmata)
                        if (lemma.GetTaskAssignment(aO.task).IndCases.Count == 0)
                            throw new Exception("Lemma "+lemma.name+" is useful for task "+aO.task+" with induction config but don't knows to which induction case it is useful.");
                    aO.findIPA("IH").Assessment = indConf.indHypConfig.maxPt * Bewertung(ind.indHyps);
                    foreach (InductionCase ic in ind.indCases)
                    {
                        bool err33 = false;
                        foreach (Error err in ic.errors)
                            if (err.errorCode == 33) { err33 = true; break; } // invalid ind cases are not assessed
                        if (err33) continue;
                        if (ic.Constructor == null) throw new Exception("Constructor of valid ind case is null in assessment!");
                        double errorPoints = ic.proof.transformationSteps.Count - Bewertung(ic);
                        double abstrMaxPt = ic.proof.transformationSteps.Count;
                        foreach(Lemma lemma in usefulLemmata)
                            if (lemma.GetTaskAssignment(aO.task).IndCases.Contains(ic.Constructor))
                            {
                                errorPoints += ( aO.AbstractMaxLemmaPoints[lemma] - Bewertung(lemma) ) 
                                             / ( lemma.usefulFor.Count * lemma.GetTaskAssignment(aO.task).IndCases.Count );
                                abstrMaxPt += aO.AbstractMaxLemmaPoints[lemma];
                            }
                        InductionCaseConfig icConfig = indConf.GetIndCaseConfig(ic.Constructor);
                        if (icConfig == null) throw new Exception("Cannot find config for induction case " + ic.Constructor.name + ".");
                        double confPt = base.confPt(icConfig.config.maxPt, abstrMaxPt, icConfig.config.minSteps);
                        aO.findIPA("Fall " + ic.Constructor.name).Assessment = Math.Max(0.0, confPt - icConfig.config.maxPt * errorPoints / Math.Max(1.0, Math.Min(icConfig.config.maxSteps, abstrMaxPt)));
                    }
                    foreach (InductionPartAssessment ipa in aO.confIndAssessment)
                        aO.assessment += ipa.Assessment;
                    if (containsError23(mainLemma)) aO.assessment -= aO.task.config.MaxPt() * weights[5];
                }
                else if (aO.task.config is SingleConfig sConf) // Assessment with simple config
                {
                    double errorPoints = 0.0;
                    foreach (Lemma lemma in usefulLemmata)
                        errorPoints += ( aO.AbstractMaxLemmaPoints[lemma] - Bewertung(lemma) ) / lemma.usefulFor.Count;
                    errorPoints = errorPoints / Math.Max(1.0, Math.Min(sConf.maxSteps, aO.AbstractMaxTaskPoints()));
                    if (containsError23(mainLemma)) errorPoints += weights[5];
                    double confPt = base.confPt(aO.task.config.MaxPt(), aO.AbstractMaxTaskPoints(), sConf.minSteps);
                    aO.assessment = Math.Max(0.0, confPt - (aO.task.config.MaxPt() * errorPoints));
                }                
            }
            bool containsError23(Lemma lemma)
            {
                foreach (Error err in lemma.errors)
                {
                    switch (err.errorCode)
                    {
                        case 20: break;
                        case 22: break;
                        case 23: return true;
                        default: throw new Exception("Found unexpected Error " + err.errorCode + " in lemma.");
                    }
                }
                return false;
            }
        }

        /// <summary>
        /// Abstract Assessment of one lemma.
        /// </summary>
        /// <param name="lemma"></param>
        /// <returns></returns>
        private double Bewertung(Lemma lemma)
        {
            foreach (Error err in lemma.errors)
                switch (err.errorCode)      
                {
                    case 20: return 0.0;
                    case 22: throw new Exception("Found error 22 in assessment of Lemmata. Lemmata with this error should not be assessed here.");
                    case 23: /*This must be done on Assessment for whole task*/break; 
                    default: throw new Exception("Found unexpected Error " + err.errorCode + " in lemma.");
                }
            if (!(bool)lemma.IsProved) return 0.0;
            return Bewertung(lemma.proof);
        }

        /// <summary>
        /// Abstract Assessment of one single proof.
        /// </summary>
        /// <param name="proof"></param>
        /// <returns></returns>
        private double Bewertung(SingleProof proof)
        {
            if (proof is Transformation tr) return Bewertung(tr);
            else if (proof is Induction ind) return Bewertung(ind);
            else throw new Exception("Found a proof that is neither Transformation nor Induction.");
        }

        /// <summary>
        /// Abstract Assessment of one Induction.
        /// </summary>
        /// <param name="ind"></param>
        /// <returns></returns>
        private double Bewertung(Induction ind)
        {
            if (HasIndErrors(ind)) return 0.0;
            double result = 0.0;
            result += Bewertung(ind.indHyps);
            foreach (InductionCase ic in ind.indCases)
                result += Bewertung(ic);
            return result;
        }

        /// <summary>
        /// Checks if there are errors directly in the induction.
        /// </summary>
        /// <param name="ind"></param>
        /// <returns> False if no errors. True if assessment must be 0.</returns>
        private bool HasIndErrors(Induction ind)
        {
            foreach (Error err in ind.errors)
                switch (err.errorCode)
                {
                    case 30: return true;
                    case 31: return true;
                    case 32: break; // nothing to do here: points of ic will be not added to result by ic is not in ind.indCases
                    case 38: return true;
                    default: throw new Exception("Found unexpected Error " + err.errorCode + " in the induction.");
                }
            return false;
        }

        /// <summary>
        /// Abstract Assessment of the set of Inductionhypothesis of one Induction.
        /// </summary>
        /// <param name="ih"></param>
        /// <returns></returns>
        private double Bewertung(List<InductionHypothesis> ihs)
        {
            double result = 0.0;
            foreach (InductionHypothesis ih in ihs)
            {
                double ihPt = 1.0;
                foreach (Error err in ih.errors)
                    switch (err.errorCode)
                    {
                        case 34: ihPt = 0.0; break;
                        case 35: ihPt = 0.0; break;
                        case 40: ihPt -= weights[6]; break;
                        case 42: break;
                        case 43: break;
                        default: throw new Exception("Found unexpected Error " + err.errorCode + " in the inductionhypothesis.");
                    }
                result += Math.Max(0.0, ihPt);
            }
            return result / ihs.Count;
        }
        /// <summary>
        /// Abstract Assessment of one Inductioncase.
        /// </summary>
        /// <param name="ic"></param>
        /// <returns></returns>
        private double Bewertung(InductionCase ic)
        {
            double errInhPt = 0.0;
            bool formalErr = false;
            int maxPt = ic.proof.transformationSteps.Count;
            foreach (Error err in ic.errors)
                switch (err.errorCode)
                {
                    case 33: formalErr = true; break;
                    case 34: formalErr = true; break;
                    case 35: formalErr = true; break;
                    case 40: errInhPt = maxPt * weights[6]; break;
                    case 41: errInhPt = maxPt * weights[7]; break;
                    default: throw new Exception("Found unexpected Error " + err.errorCode + " in the inductioncase.");
                }
            double result = 0.0;
            if (!formalErr) result += Bewertung(ic.proof) - errInhPt;
            return Math.Max(0.0, result);
        }


        /// <summary>
        /// Abstract Assessment of one Transformation.
        /// </summary>
        /// <param name="tr"></param>
        /// <returns></returns>
        private double Bewertung(Transformation tr)
        {
            foreach(Error err in tr.errors)
                switch (err.errorCode)
                {
                    case 1: return 0.0;
                    case 2: return 0.0;
                    case 39: return 0.0;
                    default: throw new Exception("Found unexpected Error " + err.errorCode + " in the transformation.");
                }
            double result = 0.0;
            foreach(TransformationStep ts in tr.transformationSteps)
            {
                bool foundError3 = false, foundError4 = false;
                foreach (Substitution sub in ts.substitutions)
                    foreach (Error err in sub.errors)
                        switch (err.errorCode)
                        {
                            case 3: foundError3 = true; break;
                            case 4: foundError4 = true; break;
                            case 13: break;
                            default: throw new Exception("Found unexpected Error " + err.errorCode + " in a Substitution.");
                        }
                if (foundError3 || foundError4) continue;
                double tsPoints = 1.0;
                foreach(Error err in ts.errors)
                    switch (err.errorCode)
                    {
                        case  5: tsPoints = 0.0; break;
                        case  6: tsPoints = 0.0; break;
                        case  7: tsPoints -= weights[1]; break;
                        case  8: tsPoints -= weights[2]; break;
                        case  9: tsPoints -= weights[3]; break;
                        case 10: tsPoints -= weights[3]; break;
                        case 11: tsPoints -= weights[4]; break;
                        case 12: tsPoints -= weights[4]; break;
                        case 21: tsPoints = 0.0; break;
                        case 36: tsPoints = 0.0; break;
                        case 37: tsPoints = 0.0; break;
                        default: throw new Exception("Found unexpected Error " + err.errorCode + " in a transformationStep.");
                    }               
                result += Math.Max(0.0,tsPoints);
            }
            return result;
        }
    }
}